/**
 * ApplicationNode.cpp
 *
 *  Created on: Feb 9, 2009
 *      Author: barrygr
 */

#include "ApplicationNode.h"

#define DATABASE_IP "217.151.194.60"
#define DATABASE_NAME "forssim"
#define FTP_IP "217.151.194.60"

using namespace FS;

/// Add this node to the H3DNodeDatabase system.
H3D::H3DNodeDatabase ApplicationNode::database( "ApplicationNode",
		&(newInstance<ApplicationNode>),
		typeid( ApplicationNode ),
		&X3DChildNode::database );

namespace ApplicationNodeInternals
{
}

ApplicationNode::ApplicationNode(int argc=1, char *argv[]=0)
: QApplication(argc, argv),
isMirroredKey("imageSettings/isMirrored"),
isStereoKey("imageSettings/isStereo"),
logDirectoryKey("logging/directory"),
caseDirectoryKey("case/directory"),
previousX3dFilePathKey("previousX3dFile/filePath"),
removeWisdomToothX3dFilePathKey("removeWisdomTooth/filePath"),
databaseNameKey("database/databaseName"),
databaseIPKey("database/databaseIP"),
databasePasswordKey("database/databasePassword"),
ftpUserKey("ftp/ftpUser"),
ftpIPKey("ftp/ftpIP"),
ftpPasswordKey("ftp/ftpPassword"),
forsDatabase(QSqlDatabase::addDatabase("QMYSQL")),
userId(-1),
forsCaseVersionId(-1),
myCDcommand(-1),
ftpAborted(false),
ftpFinished(true),
loginPressedOk(false),
isMirrored(false),
isStereo(false),
isPlayback(false),
volumeModel(NULL),
materialSegmentationModel(NULL),
forbiddenSegmentationModel(NULL),
loggingNode(NULL),
drillingStepsNode(NULL)
{
	this->setOverrideCursor(QCursor(Qt::WaitCursor));

	type_name = "ApplicationNode";
	database.initFields( this );

	// used for saving system settings
	setOrganizationName("Forsslund_Systems_AB");
	setOrganizationDomain("forsslundsystems.se");
	setApplicationName("FsWisdom");
	settings.reset(new QSettings(organizationName(), applicationName()));

	// reads the value from the settings file for mirrored. If the value isn't set, then we set it to false
	isMirrored = settings->value(isMirroredKey, false).toBool();
	settings->setValue(isMirroredKey, isMirrored);

	isStereo = settings->value(isStereoKey, false).toBool();
	settings->setValue(isStereoKey, isStereo);

	// the location of the last launched x3d file via the "Launch x3d File" tab
	QVariant variant = settings->value(previousX3dFilePathKey);
	if(variant.isNull())
		previousX3dFilePath = "";
	else
		previousX3dFilePath = variant.toString();

	// the location of the remove wisdom tooth x3d file
	variant = settings->value(removeWisdomToothX3dFilePathKey);
	if(variant.isNull())
		removeWisdomToothX3dFilePath = "";
	else
		removeWisdomToothX3dFilePath = variant.toString();

	// set where all the user logging catalogs are stored
	variant = settings->value(logDirectoryKey);

	// not currently saved in settings. a default directory therefore required. This is located in home/user and is called logDirectory
	if(variant.isNull())
		logDirectory = QDir::home().path() + QDir::separator() + "logDirectory";
	else
		logDirectory = variant.toString();

	// create the folder if it doesn't exist
	if(!logDirectory.exists())
	{
		if(!logDirectory.mkpath(logDirectory.path()))
		{
			stringstream failStringStream;
			failStringStream << "The folder: " << logDirectory.path().toStdString() <<" could not be created";
			throw ApplicationNodeException(failStringStream.str());
		}
	}
	settings->setValue(logDirectoryKey, logDirectory.path());

	// set where all the cases  are stored
	variant = settings->value(caseDirectoryKey);

	// not currently saved in settings. a default directory therefore required. This is located in home/user and is called caseDirectory
	if(variant.isNull())
		caseDirectory = QDir::home().path() + QDir::separator() + "caseDirectory";
	else
		caseDirectory = variant.toString();

	// create the folder if it doesn't exist
	if(!caseDirectory.exists())
	{
		if(!caseDirectory.mkpath(caseDirectory.path()))
		{
			stringstream failStringStream;
			failStringStream << "The folder: " << caseDirectory.path().toStdString() <<" could not be created";
			throw ApplicationNodeException(failStringStream.str());
		}
	}
	settings->setValue(caseDirectoryKey, caseDirectory.path());

	settings->sync();	// writes to the settings file direct

	// handle escape on second monitor
	keySensor.reset(new KeySensor);
	quitApplicationField.reset(new QuitApplicationField);
	keySensor->keyPress->route( quitApplicationField );

	// Show the window where the user can register themselves
	registerUserWindow.reset(new RegisterUserWindow(this));
	registerUserWindow->setGeometry(QApplication::desktop()->availableGeometry(0));
	registerUserWindow->showMaximized();

	/*Settings in config-file regarding the database.*/
	databaseName = settings->value(databaseNameKey, "").toString();
	if (databaseName == "")
	{
		settings->setValue(databaseNameKey, DATABASE_NAME);
		databaseName = DATABASE_NAME;
	}

	databaseIP = settings->value(databaseIPKey, "").toString();
	if (databaseIP == "")
	{
		settings->setValue(databaseIPKey, DATABASE_IP);
		databaseIP = DATABASE_IP;
	}
/*	databasePassword = settings->value(databasePasswordKey, "").toString();
	settings->setValue(databasePasswordKey, databasePassword);*/

	cout << "databaseIP: " << databaseIP.toStdString() << endl;
	cout << "databaseName: " << databaseName.toStdString() << endl;

	ftpIP = settings->value(ftpIPKey, "").toString();
	if (ftpIP == "")
	{
		settings->setValue(ftpIPKey, FTP_IP);
		ftpIP = FTP_IP;
	}
	/*Create ftp object. Connection is done in RegisterUser after login.*/
	ftp.reset(new QFtp(this));
	connect(ftp.get(), SIGNAL(commandFinished(int, bool)),
			this, SLOT(ftpCommandFinished(int, bool)));

	this->restoreOverrideCursor();
}

ApplicationNode::~ApplicationNode()
{
	cout << "Destroying ApplicationNode" <<endl;

	if(glWindow.get())
		glWindow->release();

	cout << "Destroyed ApplicationNode" <<endl;
}

void ApplicationNode::logout()
{
	// Hide the glWindow if it is open
	if(glWindow.get())
		glWindow->close();

	// return back to the screen where the user enters their name
	registerUserWindow->showFullScreenAndClear();

	selectCaseWindow->close();
	//Windows fix
	ftp.reset();
}

QString ApplicationNode::hashUsername(QString original)
{
      string cr;

	  cr = original.toStdString();

	  unsigned char buf[MD5_DIGEST_LENGTH];

	  MD5((unsigned char *)cr.c_str(), cr.length(), buf);

	  std::ostringstream stm;

	  for (int i=0;i<MD5_DIGEST_LENGTH;i++)
		  stm << hex << setfill('0') << setw(2) << nouppercase << (unsigned short)buf[i];


	 return QString::fromStdString(stm.str());
}



void ApplicationNode::RegisterUser(QString username, QString password)
{
	// if ftp transfer already been canceled, don't start again
	if(ftpAborted)
		return;

	this->setOverrideCursor(QCursor(Qt::WaitCursor));

	this->username = username;
	this->password = password;

	cout << "username: " << username.toStdString() << endl;
	QString temp = hashUsername(username+password);
	QString hashUsername = temp.left(16);
	QString hashPassword = temp.right(16);

	/* Maybe need to connect to database and make a query instead. It could be other problems than
	 * incorrect user name and password.
	 */
	forsDatabase.setHostName(databaseIP);
	forsDatabase.setDatabaseName(databaseName);
	forsDatabase.setUserName(hashUsername);
	forsDatabase.setPassword(hashPassword);
	if(!forsDatabase.open())
	{
		throw ApplicationNodeException("Please check user name and password.");
	}
	else
		forsDatabase.close();


	userLoggingDirectory.reset(new QDir( logDirectory.path() + "/" + username));
	// create the user's logging directory if it doesn't already exist
	if(!userLoggingDirectory->exists())
		if(!logDirectory.mkdir(username))
		{
			stringstream failStringStream;
			failStringStream << "The folder: " << userLoggingDirectory->path().toStdString() <<" cannot be created";
			throw ApplicationNodeException(failStringStream.str());
		}

	// Create user in database and get their id. If this fails, then the userId is -1
	if(forsDatabase.open())
	{
		QSqlQuery query;

		// read user from database
		query.prepare("select userId from user where name = :name;");
		query.bindValue(":name",username);

		if(query.exec())
		{
			if(query.next())
			{
				this->userId = query.value(0).toInt();
			}
		}
		else
			throw ApplicationNodeException("Failed to read user name from database");
	}

	if (!ftp.get())
	{
		//Fix because on windows it was not possible to login-logout-login again
		//without do ft.reset()...
		ftp.reset(new QFtp(this));
		connect(ftp.get(), SIGNAL(commandFinished(int, bool)),
				this, SLOT(ftpCommandFinished(int, bool)));
	}

	/*Login is ok! Download usercases over ftp.*/
	ftp->connectToHost(ftpIP);
	ftp->login(hashUsername,hashPassword);
	myCDcommand = ftp->cd("cases");

	/*Ftp download of cases is started when cd command is finished.*/
	/*On ftpDownloadComplete selectCaseWindow is opened.*/
}

bool ApplicationNode::getIsMirrored()
{
	return isMirrored;
}

void ApplicationNode::setIsMirrored(bool isMirrored)
{
	this->isMirrored = isMirrored;
	// save to settings location - registry, file etc
	settings->setValue(isMirroredKey, isMirrored);
}

bool ApplicationNode::getIsStereo()
{
	return isStereo;
}

void ApplicationNode::setIsStereo(bool isStereo)
{
	this->isStereo = isStereo;
	// save to settings location - registry, file etc
	settings->setValue(isStereoKey, isStereo);
}

QString ApplicationNode::getUsername()
{
	return username;
}

const QDir& ApplicationNode::getLogDirectory()
{
	return logDirectory;
}


const QDir& ApplicationNode::getPreviousX3dFilePath()
{
	return previousX3dFilePath;
}

void ApplicationNode::setPreviousX3dFilePath(const QDir& previousX3dFilePath)
{
	this->previousX3dFilePath = previousX3dFilePath;
	// save to settings location - registry, file etc
	settings->setValue(previousX3dFilePathKey, previousX3dFilePath.path());
}

void ApplicationNode::setRemoveWisdomToothX3dFilePath(const QDir& x3dFilePath)
{
	this->removeWisdomToothX3dFilePath = x3dFilePath;
	// save to settings location - registry, file etc
	settings->setValue(removeWisdomToothX3dFilePathKey, removeWisdomToothX3dFilePath.path());
}

const QDir& ApplicationNode::getCaseDirectory() const
{
	return caseDirectory;
}


void ApplicationNode::exitX3DWithStandardGuiWindow()
{
	if( glWindow.get() )
		glWindow->stopTimer();

	cout << "exitX3DWithStandardGuiWindow" <<endl;
	if (this->isPlayback)
	{
		switch(QMessageBox::question(NULL,tr("Do you want to Exit"),
				tr("Do you want to Exit?"),
				QMessageBox::Yes|QMessageBox::Cancel,QMessageBox::Yes))
		{
		case QMessageBox::Yes:
			// back to select case menu
			selectCaseWindow->setGeometry(QApplication::desktop()->availableGeometry(0));
			selectCaseWindow->showFullScreen();
			CleanupX3dWindows();

			return;
		default:
			if( glWindow.get() )
				glWindow->startTimer();

			return;
		}
	}
	else
	{
		switch(QMessageBox::question(NULL,tr("Save Data on Exit"),
				tr("Do you wish to save the data associated with this case?"),
				QMessageBox::Yes|QMessageBox::No|QMessageBox::Cancel,QMessageBox::Yes))
		{
		case QMessageBox::Yes:
			// save timestamps
			this->setOverrideCursor(QCursor(Qt::WaitCursor));
			volumeModel = static_cast <VolumeModel*> (glWindow->getSceneNode("VM"));
			materialSegmentationModel = static_cast< MaterialSegmentationModel * >(glWindow->getSceneNode("SM"));

			//Number of Forbidden voxels
			forbiddenSegmentationModel = static_cast< ForbiddenSegmentationModel * >(glWindow->getSceneNode("FM"));

			//Number of expert voxels
			drillingStepsNode = static_cast< DrillingStepsNode * >(glWindow->getSceneNode("DS"));

			//Done below instead
			//volumeModel->saveTimeStamps();
			//Not needed anymore, data is are sent to database instead.
			//volumeModel->saveStatistics();

			// Save user case to database
			if(forsDatabase.open())
			{
				QSqlQuery query;

				query.prepare("INSERT INTO userCase (userId, forsCaseVersionId, timeStampFile) "
						"VALUES (:userId, :forsCaseVersionId, :timeStampFile)");
				query.bindValue(":userId", userId);
				query.bindValue(":forsCaseVersionId", forsCaseVersionId);
				QFile file( QDir(  QString (volumeModel->saveTimeStamps().c_str()) ).path() );
				file.open(QIODevice::ReadOnly);
				QByteArray byteArray = file.readAll();
				query.bindValue(":timeStampFile", byteArray);

				if(query.exec())
					cout << "Added recorded data to database" << endl;
				else
					throw ApplicationNodeException("Failed to add recorded data to database");

				MFUnsignedInt8::const_iterator segmentIdFieldIterator;

				query.prepare("SELECT LAST_INSERT_ID()");
				if(!query.exec())
					throw ApplicationNodeException("Failed to read last index from database");
				query.next();
				int index = query.value(0).toInt();

				for(int drillingStep=1;drillingStep<=(int)drillingStepsNode->getNoOfDrillingSteps();drillingStep++){
					for(segmentIdFieldIterator=forbiddenSegmentationModel->segmentIdField->begin();
					segmentIdFieldIterator<forbiddenSegmentationModel->segmentIdField->end();
					segmentIdFieldIterator++)
					{
						query.prepare("INSERT INTO userSegments (userCaseId, segmentId, drillingStep, voxelsRemoved, noVoxelsEqualToExpert) "
								"VALUES (:userCaseId, :segmentId, :drillingStep, :voxelsRemoved, :noVoxelsEqualToExpert)");
						query.bindValue(":userCaseId", index );
						query.bindValue(":segmentId", -(*segmentIdFieldIterator));
						query.bindValue(":drillingStep", drillingStep);
						query.bindValue(":voxelsRemoved", forbiddenSegmentationModel->noOfForbiddenVoxelsDrilledByUser((*segmentIdFieldIterator),drillingStep));
						query.bindValue(":noVoxelsEqualToExpert", 0);

						if(query.exec())
							cout << "Added forbidden segment data to database " <<query.lastError().databaseText ().toStdString()<< endl;
						else
							cout << "No forbidden segments added to database" << query.lastError().databaseText ().toStdString()<< endl;
					}

					int noOfSegments = materialSegmentationModel->getNoOfSegments();
					for (int i = 1;i<noOfSegments;i++){
						query.prepare("INSERT INTO userSegments (userCaseId, segmentId, drillingStep, voxelsRemoved, noVoxelsEqualToExpert) "
								"VALUES (:userCaseId, :segmentId, :drillingStep, :voxelsRemoved, :noVoxelsEqualToExpert)");
						query.bindValue(":userCaseId", index );
						query.bindValue(":segmentId", i);
						query.bindValue(":drillingStep", drillingStep);
						query.bindValue(":voxelsRemoved", drillingStepsNode->noOfNonExpertVoxelsDrilledByUser(i,drillingStep) + drillingStepsNode->noOfExpertVoxelsDrilledByUser(i,drillingStep));
						query.bindValue(":noVoxelsEqualToExpert", drillingStepsNode->noOfExpertVoxelsDrilledByUser(i,drillingStep));

						if(query.exec()){
							cout << "Added segment data to database" << query.lastError().databaseText ().toStdString()<< endl;
						}
						else{
							cout << "No segments added to database" << query.lastError().databaseText ().toStdString()<< endl;
						}
					}
				}
			}

			this->restoreOverrideCursor();
			// falls down into no
		case QMessageBox::No:
			// back to select case menu
			selectCaseWindow->setGeometry(QApplication::desktop()->availableGeometry(0));
			selectCaseWindow->showFullScreenAndUpdate();
			CleanupX3dWindows();

			return;
		default:
			glWindow->startTimer();
			return;
		}
	}
}

bool ApplicationNode::CleanupX3dWindows()
{
	if( !glWindow.get() )
		return false;

	if( !removeWisdomToothWindow.get() )
		return false;

	//Destroy the RemovewisdomToothwindow and clear the autoptr
	removeWisdomToothWindow->DestroyRemoveWisdomToothWindow();
	removeWisdomToothWindow.reset();

	//Destroy the openGL window and all its data. Clear the autoptr
	glWindow->DestroyOpenGLWindow();
	glWindow.reset();

	return true;
}

void ApplicationNode::launchX3D(const QDir& x3dFilePath)
{
	if(!glWindow.get())
		// create and launch the gl window
		glWindow.reset(new GlWindowNode(this, x3dFilePath));
	else
		glWindow->launchX3D(x3dFilePath);
}

void ApplicationNode::launchX3DWithStandardGuiWindow(const QDir& x3dFilePath, bool playback, int forsCaseVersionId)
{
	this->forsCaseVersionId = forsCaseVersionId;
	QFileInfoList fileInfoList = userLoggingDirectory->entryInfoList ( QDir::Files, QDir::Time );

	QDir playbackFilePath;
	if(playback)
	{
		// find the last playback file saved for this user
		bool foundPlaybackFileForUser = false;
		QFileInfoList::const_iterator i;
		for(i=fileInfoList.begin();i<fileInfoList.end();i++)
			if((*i).suffix()=="nrrd")
			{
				playbackFilePath = (*i).absoluteFilePath();
				foundPlaybackFileForUser = true;
				break;
			}

		// nothing to playback
		if(!foundPlaybackFileForUser)
			return;
	}

	if(!glWindow.get())
		//PH	glWindow.reset(new GlWindowNode(this, removeWisdomToothX3dFilePath));
		glWindow.reset(new GlWindowNode(this, x3dFilePath));
	else
		//PH		glWindow->launchX3D(removeWisdomToothX3dFilePath);
		glWindow->launchX3D(x3dFilePath);

	// show the gui screen
	if(!removeWisdomToothWindow.get())
		removeWisdomToothWindow.reset(new RemoveWisdomToothWindow(this, playback));
	else
		removeWisdomToothWindow->initialize(playback);

	volumeModel 	= static_cast <VolumeModel*>	(glWindow->getSceneNode("VM"));
	loggingNode		= static_cast <LoggingNode*>	(glWindow->getSceneNode("LN"));

	// set where we are going to log
	if(loggingNode)
		loggingNode->userLoggingPathField->setValue(userLoggingDirectory->path().toStdString());

	this->isPlayback = false;
	// set playback data
	if(playback)
	{
		this->isPlayback = true;
		if(volumeModel)
			volumeModel->setTimeStampFile(playbackFilePath.path().toStdString());
	}

	removeWisdomToothWindow->setGeometry(QApplication::desktop()->availableGeometry(0));
	removeWisdomToothWindow->showFullScreen();
	selectCaseWindow->close();

}

Node* ApplicationNode::getSceneNode(const string& nodeName)
{
	return glWindow->getSceneNode(nodeName);
}

void ApplicationNode::setWaitCursor()
{
	this->setOverrideCursor(QCursor(Qt::WaitCursor));
}

void ApplicationNode::restoreNormalCursor()
{
	this->restoreOverrideCursor();
}

void ApplicationNode::ftpCommandFinished(int commandID, bool error)
{
	cout << "ApplicationNode ftpCommandFinished: " << endl;

	cout << "state: " << ftp->state() << endl;

	if (ftp->currentCommand() == QFtp::ConnectToHost) {
		if (error) {
			std::cout << "CommandError ConnectToHost" << std::endl;
			return;
		}
		std::cout << "ApplicationNode CommandFinished ConnectToHost" << std::endl;
		return;
	}

	if (ftp->currentCommand() == QFtp::Login) {
		if (error) {
			return;
		}
		std::cout << "ApplicationNode CommandFinished Login" << std::endl;
		return;
	}

	if (ftp->currentCommand() == QFtp::Cd) {
		if (error) {
			std::cout << "ApplicationNode CommandError CD" << std::endl;
			return;
		}
		std::cout << "ApplicationNode CommandFinished CD" << std::endl;
		if (commandID == myCDcommand)
		{
			progressDialog.reset(new ProgressDialog(this,"-Welcome to FsWisdom-\n Downloading and installing cases!","Cancel",0,5));
			QObject::connect(this->getProgressDialog(),SIGNAL(canceled()),this,SLOT(cancelFtp()));
			progressDialog->show();
			ftpFinished = false;

			/*Create localCaselist and serverCaseList and download cases over ftp if not already downloaded.*/
			localCaseList.reset(new LocalCaseList(this,this->getCaseDirectory(), this->getFtp()));
			QObject::connect(this->getLocalCaseList(), SIGNAL(downloadComplete()),
					this, SLOT(ftpDownloadComplete()));
			serverCaseList.reset(new ServerCaseList(this->getForsDatabase()));
			serverCaseList->syncWithLocal(this,*localCaseList.get());
		}
		return;
	}
}

void ApplicationNode::ftpCommandStarted(int id)
{
	std::cout << "ApplicationNode CommandStarted: " << id << std::endl;
}

void ApplicationNode::ftpDownloadComplete()
{
	cout << "ApplicationNode::ftpDownloadComplete()" << endl;
	// Show the window where the user can register themselves
	/*registerUserWindow.reset(new RegisterUserWindow(this));
	registerUserWindow->setGeometry(QApplication::desktop()->availableGeometry(0));
	registerUserWindow->showMaximized();*/
	ftpFinished = true;
	ftp->close();
	if (progressDialog.get() != NULL)
		progressDialog->reset();

	// show the select case screen
	if(!selectCaseWindow.get())
		selectCaseWindow.reset(new SelectCaseWindow(this));
	else
		selectCaseWindow->initialize();

	//Moved to showSelectCaseWindow() so can wait until all cases are shown.
	selectCaseWindow->setGeometry(QApplication::desktop()->availableGeometry(0));
	selectCaseWindow->showFullScreen();

	// finished registering the user
	registerUserWindow->close();
	loginPressedOk = false;

	this->restoreOverrideCursor();
}

QFtp* ApplicationNode::getFtp()
{
	return ftp.get();
}

LocalCaseList* ApplicationNode::getLocalCaseList()
{
	return localCaseList.get();
}

ProgressDialog* ApplicationNode::getProgressDialog()
{
	return progressDialog.get();
}

void ApplicationNode::setFtpAborted(bool a_aborted)
{
	ftpAborted = a_aborted;
	if(a_aborted)
		ftpFinished = true;
}

bool ApplicationNode::getFtpAborted()
{
	return ftpAborted;
}

void ApplicationNode::cancelFtp()
{
	this->getProgressDialog()->cancelFtpDownloading();
}

void ApplicationNode::setLoginPressedOk(bool pressed)
{
	loginPressedOk = pressed;
}

bool ApplicationNode::getLoginPressedOk()
{
	return loginPressedOk;
}

